-- An implimentation of the Traveling Salesman Problem
--
-- Author	Matthew Caryl
-- Created	15.11.96

with Ada.Text_IO, Ada.Command_Line, Ada.Text_IO.Integer_IO, Ada.Strings, Ada.Unchecked_Deallocation, Ada.Numerics.Elementary_Functions, Matrix;
use Matrix, Ada.Text_IO;

procedure Salesman is
	type Long is new Long_Integer;
	type Percentage is range 0..100;
	type Chromosome is array (Integer range <>) of Integer;
	type Chromosome_Access is access Chromosome;
	type Population is array (Integer range <>) of Chromosome_Access;
	type Evaluations is array (Integer range <>) of Integer;
	
	Input_Error : exception;

	seed : Long;

	function Random(n : in Integer) return Integer is
	begin
	end Random;

	function Chromo(n : in Integer) return Chromosome_Access is
		c : Chromosome_Access;
		t, p : Integer;
	begin
		c := new Chromosome(1..n);
		for i in 1..n loop
			c(i) := i;
		end loop;
		for i in 1..n loop
			t := c(i);
			p := Random(n) + 1;
			c(i) := c(p);
			c(p) := t;
		end loop;
		return c;
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("Initialise");
			raise;
	end Chromo;

	function Chromo(p : in Chromosome_Access) return Chromosome_Access is
		c : Chromosome_Access;
		m1, m2 : Integer;
	begin
		c := new Chromosome(1..p'Length);

		-- two random positions, m1 < m2
		m1 := Random(p'Length - 1) + 1;
		m2 := Random(p'Length - m1) + m1 + 1;

		-- copy positions before m1
		c(1..(m1 - 1)) := p(1..(m1 - 1));

		-- place contents of m2 before m1
		c(m1) := p(m2);
		c(m1 + 1) := p(m1);

		-- copy positions between m1 and m2
		c((m1 + 2)..m2) := p((m1 + 1)..(m2 - 1));

		-- copy positions after m2
		c((m2 + 1)..p'Length) := p((m2 + 1)..p'Length);

		return c;
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("Mutate");
			raise;
	end Chromo;
	
	function Search(c : Chromosome; v : Integer) return Boolean is
		p, q : Integer;
	begin
		p := c'First;
		q := c'Last;
		while (p /= q) loop
			if c(p) = v then
				q := p;
			else
				p := p + 1;
			end if;
		end loop;
		return c(p) = v;
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("Search");
			raise;
	end Search;

	function Chromo(p1, p2 : in Chromosome_Access) return Chromosome_Access is
		c : Chromosome_Access;
		k1, k2 : Integer;
		p, q, r : Integer;
	begin
		c := new Chromosome(1..p1'Length);

		-- two random positions, k1 < k2, k1 + k2 <= p'Length / 2
		k1 := Random(p1'Length / 2) + 1;
		k2 := Random(p1'Length - p1'Length / 2) + 1 + k1;
	
		-- set of cities within selected region are selected cities
	
		-- copy from 1st or second parent depending on whether value is in selected area
		r := k1;
		for i in 1..p1'Length loop
			if Search(p1(k1..k2), p2(i)) then 
				c(i) := p1(r);
				r := r + 1;
			else
				c(i) := p2(i);
			end if;
		end loop;
		
		return c;
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("Cross");
			raise;
	end Chromo; 

	function Evaluate(c : in Chromosome_Access; graph : Matrix.Matrix_Access) return Float is
		total : Float := 0.0;
	begin
		for i in c'First..(c'Last - 1) loop
			total := total + Get(graph.all, c(i), c(i + 1));
		end loop;
		total := total + Get(graph.all, c'First, c'Last);
		return total;
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("Evaluate");
			raise;
	end Evaluate;

	function AntiRoulette(evals : Evaluations; total : Integer; max : Integer) return Integer is
		n : Integer;
		sum : Integer;
		i : Integer;
	begin
		n := Random(total);
		sum := 0;
		i := 1;
		while n >= sum + evals(i) loop
			sum := sum + evals(i);
			i := i + 1;
		end loop;
		return i;
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("AntiRoulette");
			raise;
	end AntiRoulette;
	
	package Integer_IO is new Ada.Text_IO.Integer_IO(Integer);
	package Float_IO is new Ada.Text_IO.Float_IO(Float);
	package Long_IO is new Ada.Text_IO.Integer_IO(Long);
	package Percentage_IO is new Ada.Text_IO.Integer_IO(Percentage);

	use Ada.Text_IO, Ada.Command_Line;
	procedure Error is
	begin
		Put_Line("Error: incorrect parameters");
		Put_Line("Correct format: " & Command_Name & " (source name) (optimum name) (population) (generation) (mutation) (crossover) (seed)");
	end Error;

	use Ada.Text_IO, Ada.Command_Line;
	function InputSourceName return String is
		last : Positive;
	begin
		if Argument_Count /= 7 then
			raise Input_Error;
		end if;
		return Argument(1);
	exception
		when Constraint_Error | Input_Error =>
			Error;
			raise;
	end InputSourceName;
	
	use Ada.Text_IO, Ada.Command_Line;
	function InputOptimumName return String is
		last : Positive;
	begin
		if Argument_Count /= 7 then
			raise Input_Error;
		end if;
		return Argument(2);
	exception
		when Constraint_Error | Input_Error =>
			Error;
			raise;
	end InputOptimumName;

	use Ada.Text_IO, Ada.Command_Line, Percentage_IO, Long_IO, Integer_IO;
	procedure Input(population, generation : out Integer; mutation, crossover : out Percentage; seed : out Long) is
		last : Positive;
	begin
		if Argument_Count /= 7 then
			raise Input_Error;
		end if;
		Get(Argument(3), population, last);
		Get(Argument(4), generation, last);
		Get(Argument(5), mutation, last);
		Get(Argument(6), crossover, last);
		Get(Argument(7), seed, last);
	exception
		when Constraint_Error | Input_Error =>
			Error;
			raise;
	end Input;
	
	use Ada.Text_IO, Integer_IO;
	procedure Output(pop : in Population; c : Chromosome_Access; name : in String; graph : in Matrix.Matrix_Access) is
		file : File_Type;
	begin
		Create(file, Out_File, name);
		for i in pop'Range loop
			Put(file, Integer(Evaluate(pop(i), graph)));
			New_Line(file);
		end loop;
		New_Line(file);
		Put(file, Integer(Evaluate(c, graph)));
		Close(file);
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("Output");
			raise;
	end Output;

	function Sqr(x : in Float) return Float is
	begin
		return x * x;
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("Sqr");
			raise;
	end Sqr;
	
	use Ada.Text_IO, Float_IO, Ada.Numerics.Elementary_Functions;
	procedure ReadData(name : in String; graph : out Matrix_Access; size : out Integer) is
		type Coord is
			record
				x, y : Float;
			end record;
		type Cities is array(Integer range <>) of Coord;
		type Cities_Access is access Cities;
		file : File_Type;
		c : Cities_Access;
		dummy : Integer;
	begin
		Open(file, In_File, name);
		Skip_Line(file);
		Skip_Line(file);
		Skip_Line(file);
		Set_Col(file, 11);
		Get(file, size);
		graph := new Matrix.Matrix(size, size);
		Skip_Line(file);
		Skip_Line(file);
		Skip_Line(file);
		Skip_Line(file);
		c := new Cities(1..size);
		for i in 1..size loop
			Get(file, dummy);
			Get(file, c(i).x);
			Get(file, c(I).y);
		end loop;
		for i in 1..size loop
			for j in 1..size loop
				Put(graph.all, Sqrt(Sqr(c.all(i).x - c.all(j).x) + Sqr(c.all(i).x - c.all(j).x)), i, j);
			end loop;
		end loop;
		Close(file);
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("ReadData");
			raise;
	end ReadData;
	
	use Ada.Text_IO, Integer_IO;
	function ReadSolution(name : in String) return Chromosome_Access is
		file : File_Type;
		size : Integer;
		c : Chromosome_Access;
	begin
		Open(file, In_File, name);
		Skip_Line(file);
		Skip_Line(file);
		Skip_Line(file);
		Set_Col(file, 12);
		Get(file, size);
		Skip_Line(file);
		Skip_Line(file);
		c := new Chromosome(1..size);
		for i in 1..size loop
			Get(file, c(i));
		end loop;
		Close(file);
		return c;
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("ReadSolution");
			raise;
	end ReadSolution;

	procedure Free is new Ada.Unchecked_Deallocation(Chromosome, Chromosome_Access);
	
	use Ada.Text_IO, Integer_IO;
	procedure Main(height, width : in Integer; mutation, crossover : in Percentage; source, optimum : String) is
		graph : Matrix_Access;
		pop : Population(1..height);
		offspring : Population(1..height);
		eval : Evaluations(1..height);
		keep : array (1..height) of boolean;
		size : Integer;
		total : Integer;
		r : Percentage;
		max, min : Integer;
		file : File_Type;
	begin
		Create(file, Out_File, "best.txt");
		
		Put_Line("Reading data");
		ReadData(source, graph, size);
		
		Put_Line("Initialising population");
		-- initialise
		for i in pop'Range loop
			pop(i) := Chromo(size);
		end loop;

		Put_Line("Evolving population");
		-- generation loop
		for generation in 1..width loop
			-- evaluate
			total := 0;
			for i in 1..height loop
				eval(i) := Integer(Evaluate(pop(i), graph));
				total := total + eval(i);
				keep(i) := false;
			end loop;

			-- find maximum and minimum
			max := 1;
			min := 1;
			for i in 2..height loop
				if eval(max) < eval(i) then
					max := i;
				elsif eval(min) > eval(i) then
					min := i;
				end if;
			end loop;
			
			-- record best
			Put(file, Integer(Evaluate(pop(min), graph)));
			New_Line(file);
		
			-- breed
			for i in 1..height loop
				r := Percentage(Random(100));

				if (r < mutation) then
					offspring(i) := Chromo(pop(i));
				elsif (r < mutation + crossover) then
					offspring(i) := Chromo(pop(i), pop(AntiRoulette(eval, total, max)));
				else
					offspring(i) := pop(i);
					keep(i) := true;
				end if;
			end loop;

			-- replace
			for i in 1..height loop
				if (not keep(i)) then
					Free(pop(i));
				end if;
			end loop;
			pop := offspring;
		end loop;

		Put_Line("Writing population");
		Output(pop, ReadSolution(optimum), "evals.txt", graph);

		Close(file);
	
		Put_Line("Finished");
	exception
		when CONSTRAINT_ERROR | DATA_ERROR =>
			Put_Line("main");
			raise;
	end Main;

	width : Integer;
	height : Integer;
	mutation : Percentage;
	crossover : Percentage;

begin
	Input(height, width, mutation, crossover, seed);
	Main(height, width, mutation, crossover, InputSourceName, InputOptimumName);
exception
	when Input_Error =>
		null;
	when CONSTRAINT_ERROR | DATA_ERROR =>
		Put_Line("salesman");
			raise;
end Salesman;
